/****************************<--OGI-->*************************************/
/*                                                                        */
/*             Center for Spoken Language Understanding                   */
/*        Oregon Graduate Institute of Science & Technology               */
/*                         Portland, OR USA                               */
/*                        Copyright (c) 1999                              */
/*                                                                        */
/*      This module is not part of the CSTR/University of Edinburgh       */
/*               release of the Festival TTS system.                      */
/*                                                                        */
/*  In addition to any conditions disclaimers below, please see the file  */
/*  "license_ogi_tts.txt" distributed with this software for information  */
/*  on usage and redistribution, and for a DISCLAIMER OF ALL WARRANTIES.  */
/*                                                                        */
/****************************<--OGI-->*************************************/
#include "OGI_TimeWarp.h"

float OGI_TimeWarp::invmap(float y_in, float *x_indx) {
  // returns interpolated, inverse-mapped value v(*x_indx)
  // int part of *x_indx is index to left in x array
  // float part of *x_indx is weight to apply to indx to RIGHT in x array
  // round *x_indx to get nearest int x value
  //
  // **implicitly assumes time pair (x[-1],y[-1]) = (0,0)
  //
  // Lindx is cached to try to save search time
  // N is number of src/targ segments
  float v;   // invmapped return value 
  float gR;  // floating pt part of x_indx returned

  // degenerate cases
  if ((y_in < 0) || (N < 1)){
    *x_indx = 0.;
    Lindx = 0;
    return 0;
  } 
  if (y_in <= y[0]){
    if (y[0] > 0.0)
      gR = y_in/y[0];
    else
      gR = 1.0;
    v  = gR*x[0];
    *x_indx = 0.;
    Lindx = 0;
    return v;
  }  
  if (y_in >= y[N-1]){
    v  = x[N-1];
    gR = 0.0;
    *x_indx = N-1;
    return v;
  }

  // reset Lindx if out of bounds or too large
  if ((Lindx > N-1) || (Lindx < 0))
    Lindx = 0;
  if (y[Lindx] > y_in)  
    Lindx = 0;

  // now search for correct L & R neighbors
  for ( ; Lindx<N-1; Lindx++){
    if (y[Lindx+1] > y_in)
      break;
  }

  // verify...
  if (!((y[Lindx] <= y_in) && (y_in <= y[Lindx+1]))){
    save_src("debug_OGI_Timewarp_src.phn");
    save_targ("debug_OGI_Timewarp_targ.phn");
    fprintf(stderr,"OGI_TimeWarp::invmap() problem, Lindx=%d, y_in=%f \n  dumped src and targ info to debug_OGI_TimeWarp_*.phn files.",Lindx,y_in);
    EST_error("exiting...");
  }

  gR = (y_in - y[Lindx]) / (y[Lindx+1]-y[Lindx]);
  v  =  (1-gR)*x[Lindx] + gR*x[Lindx+1];
  *x_indx = Lindx + gR;

  //  printf("invmap: y=%f ==> x=%f (x_indx=%f)\n", y_in, v,*x_indx);

  return v;
}



float OGI_TimeWarp::fwdmap(float x_in, float *y_indx) {
  // returns interpolated, fwderse-mapped value v(*y_indx)
  // int part of *y_indx is index to left in y array
  // float part of *y_indx is weight to apply to indx to RIGHT in y array
  // round *y_indx to get nearest int y value
  //
  // **implicitly assumes time pair (y[-1],x[-1]) = (0,0)
  //
  // Lindx is cached to try to save search time
  // N is number of src/targ segments
  float v;   // fwdmapped return value 
  float gR;  // floating pt part of y_indx returned

  // degenerate cases
  if ((x_in < 0) || (N < 1)){
    *y_indx = 0.;
    Lindx = 0;
    return 0;
  } 
  if (x_in <= x[0]){
    if (x[0] > 0.0)
      gR = x_in/x[0];
    else
      gR = 1.0;
    v  = gR*y[0];
    *y_indx = 0.;
    Lindx = 0;
    return v;
  }  
  if (x_in >= x[N-1]){
    v  = y[N-1];
    gR = 0.0;
    *y_indx = N-1;
    return v;
  }

  // reset Lindx if out of bounds or too large
  if ((Lindx > N-1) || (Lindx < 0))
    Lindx = 0;
  if (x[Lindx] > x_in)  
    Lindx = 0;

  // now search for correct L & R neighbors
  for ( ; Lindx<N-1; Lindx++){
    if (x[Lindx+1] > x_in)
      break;
  }

  // verify...
  if (!((x[Lindx] <= x_in) && (x_in <= x[Lindx+1]))){
    save_src("debug_OGI_Timewarp_src.phn");
    save_targ("debug_OGI_Timewarp_targ.phn");
    fprintf(stderr,"OGI_TimeWarp::fwdmap() problem, Lindx=%d, x_in=%f \n  dumped src and targ info to debug_OGI_TimeWarp_*.phn files.",Lindx,x_in);
    EST_error("exiting...");
  }

  gR = (x_in - x[Lindx]) / (x[Lindx+1]-x[Lindx]);
  v  =  (1-gR)*y[Lindx] + gR*y[Lindx+1];
  *y_indx = Lindx + gR;

  //  printf("fwdmap: x=%f ==> y=%f (y_indx=%f)\n", x_in, v,*y_indx);

  return v;
}


