/****************************<--OGI-->*************************************/
/*                                                                        */
/*             Center for Spoken Language Understanding                   */
/*        Oregon Graduate Institute of Science & Technology               */
/*                         Portland, OR USA                               */
/*                        Copyright (c) 1999                              */
/*                                                                        */
/*      This module is not part of the CSTR/University of Edinburgh       */
/*               release of the Festival TTS system.                      */
/*                                                                        */
/*  In addition to any conditions disclaimers below, please see the file  */
/*  "license_ogi_tts.txt" distributed with this software for information  */
/*  on usage and redistribution, and for a DISCLAIMER OF ALL WARRANTIES.  */
/*                                                                        */
/****************************<--OGI-->*************************************/
#include "EST.h"
#include "EST_Pathname.h"
#include "OGI_Track.h"
#include "OGI_Macro.h"
#include "OGI_Filter.h"
#include "resLPC_pmark.h"
#include "resLPC_analysis.h"

static void parse_init_params(LISP params, EST_Features &In);
static void compute_pmarks( EST_Wave &lx, EST_Features &g, OGI_Track &pmarks);
static void retune_pmarks(EST_Wave &sp, EST_Features &g, OGI_Track &pmarks);
static void compute_E(EST_Wave &sig, OGI_Track &E, EST_Features &g);
static void design_ma_filter(int N, EST_DVector &b);
static void max_limit_curve(OGI_Track &E);
static void cluster_2class(OGI_Track &fv, float Vbias, EST_FList &B, EST_FList &E);
static int find_min(EST_Wave &x, int B, int E);
static void find_pgzc(EST_Wave &x, int B, int E, EST_IList &pgzc);
static void find_vpm( EST_Wave &lx, EST_Wave &lxdiff, EST_FList &Blist, EST_FList &Elist, EST_IList &pm_cand);
static void IList2Track(EST_IList &pmList, float Fs, OGI_Track &pm);
static void purge_pm_list(EST_IList &pm, EST_IList &rmlist);
static bool rm_suspicious_pm(EST_Wave &lx, EST_IList &pmList, EST_Features &g);

static void max_V_xcorr_shift( OGI_Track &pm, EST_Wave &sig, EST_Features &g);
static void s2p(EST_FVector &frame, float &power);

#define MAX_ITER 30
bool pm_debug_print = 0;
bool pm_debug_dump  = 0;

// can use (directory-entries ".") to get file list when called in scm

LISP OGIresLPC_Pmark_Analysis(LISP params){
  EST_Features g;
  parse_init_params(params, g);
  LISP l;
  
  LISP lfiles = get_param_lisp("infile_list", params, NIL);

  pm_debug_print = (!(g.I("run_quietly")));
  pm_debug_dump  = g.I("dump_intermediate_files");

  if (lfiles == NIL){
    EST_error("OGIresLPC pmark analysis:  infile_list empty - no input files to process");
  }

  EST_Pathname f,spfile,lxfile,outfile,basename;
  EST_String sp_path  = wstrdup(g.S("sp_path"));
  EST_String lx_path  = wstrdup(g.S("lx_path"));
  EST_String out_path = wstrdup(g.S("out_path"));
  EST_String out_ext  = wstrdup(g.S("out_ext"));
  EST_String sp_ext   = wstrdup(g.S("sp_ext"));
  EST_String lx_ext   = wstrdup(g.S("lx_ext"));
  for (l=lfiles; l != 0; l=cdr(l)){
    bool files_loaded = TRUE;
    f = EST_Pathname(get_c_string(car(l)));
    basename = f.basename(1);
    lxfile  = lx_path + basename + lx_ext;
    spfile  = sp_path + basename + sp_ext;
    outfile = out_path   + basename + out_ext;
    if (pm_debug_print) cout << " loading " << lxfile << " --> output to " << outfile << endl;

    EST_Wave lx;    
    EST_Wave sp;    
    OGI_Track pmarks;
    if (lx.load(lxfile) != read_ok){
      cout << "Could not load lx file " << lxfile << "  ... skipping." << endl;
      files_loaded = FALSE;
    }
    if (sp.load(spfile) != read_ok){
      cout << "Could not load speech file " << spfile << "  ... skipping." << endl;
      files_loaded = FALSE;
    }

    if (files_loaded){
      //
      // do it
      //
      compute_pmarks(lx, g, pmarks);
      retune_pmarks( sp, g, pmarks);
      pmarks.save(outfile);
    }
  }

  return NIL;
}


//
//  Compute pitchmarks from laryngograph signal
//
static void compute_pmarks(
			   EST_Wave &lx, 
			   EST_Features &g, 
			   OGI_Track &pmarks){

  if ((g.S("lx_sign") == "down_is_closed_glottis") || (g.S("lx_sign") == "up_is_open_glottis"))
    ;// nothing
  else if ((g.S("lx_sign") == "down_is_open_glottis") || (g.S("lx_sign") == "up_is_closed_glottis")){
    invert(lx);
  }
  else 
    EST_error("pmark_analysis.cc:  don't understand your 'lx_sign setting...\n");

  EST_DVector a,b;
  if (pm_debug_dump) lx.save("lx.wav","riff");

  ////
  //// Pre-filter to get rid of low-freq drift on signal
  ////
  filtlist_to_vector(g.S("pre_filt1"), b, a);
  OGIfiltfilt(b, a, lx);
  if (pm_debug_dump) lx.save("lx_prefilt.wav","riff");
  
  ////
  //// Compute derivative of Lx
  ////
  EST_Wave lxdiff = lx;
  filtlist_to_vector(g.S("diff_filt"), b, a);
  OGIzerophase_FIRfilter(b, lxdiff);
  if (pm_debug_dump) lxdiff.save("lxdiff.wav","riff");
 
  ////
  //// Make V/UV decision
  ////
  // filter to isolate voicing band
  EST_Wave lx_vuvfilt = lx;
  filtlist_to_vector(g.S("vuv_filt"), b, a);
  OGIfiltfilt(b, a, lx_vuvfilt);
  if (pm_debug_dump) lx_vuvfilt.save("lx_vuv.wav","riff");

  // Square and do limiting on lx_vuvfilt
  OGI_Track lxE;
  compute_E(lx_vuvfilt, lxE, g);
  max_limit_curve( lxE );

  // 1 (guessed) pitch period smoother (two passes)
  float tt = g.F("Eest_ppm");
  design_ma_filter((int)(tt/2.0+0.5), b); 
  OGIzerophase_FIRfilter(b, lxE);

  // Find optimal threshold between V and UV based on energy
  EST_FList Vbegin, Vend; // in seconds
  cluster_2class(lxE,  g.F("Vbias"), Vbegin, Vend);
  if (pm_debug_dump) lxE.save("lxE.est");

  ////
  //// Find voiced pitchmarks: 
  ////    
  EST_IList pm_cand;
  find_vpm(lx_vuvfilt, lxdiff, Vbegin, Vend, pm_cand);
  if (pm_debug_dump){
    IList2Track(pm_cand, (float) lx.sample_rate(), pmarks);
    pmarks.save("pm_unmod.est");
  }

  ////
  //// Remove stray and double pitchmarks 
  ////
  bool there_are_suspects = TRUE;
  int n=0;
  while ((there_are_suspects) && (n<MAX_ITER)){
    there_are_suspects = rm_suspicious_pm(lx, pm_cand, g);
    n++;
  }

  ////
  //// Output
  ////
  IList2Track(pm_cand, (float) lx.sample_rate(), pmarks);
  if (pm_debug_dump) pmarks.save("pm_prune.est");
}




//
//  Shift to get rid of wave propagation delay by aligning 
//  to peak in LPC residual.  (Possibly prune some more around edges as well...)
//
static void retune_pmarks(
			  EST_Wave &sp, 
			  EST_Features &g, 
			  OGI_Track &pmarks){
  
  // fill in UV parts with default pmarks (to keep LPC well-behaved)
  OGI_Track lpcpm;
  lpcpm = pmarks;
  fill_UV_pmarks(lpcpm, sp.length(), sp.sample_rate(), g, 0);  

  // compute LPC
  compute_lpc(sp, lpcpm, g);

  // compute residual
  EST_Wave res;
  compute_residual(sp, lpcpm, g, res);

  // shift pmarks to max xcorr point
  max_V_xcorr_shift(pmarks, res, g);
}



static void max_V_xcorr_shift( OGI_Track &pm, EST_Wave &sig, EST_Features &g){
  
  float Fs = (float) sig.sample_rate();
  int N = g.I("shift_search_range");
  int t,t_shift, best_shift=0;
  double acc,max_acc=0.0;
  int Nsig = sig.length();

  for (int k=-N; k<=N; k++){
    t_shift = g.I("shift_guess") + k;
    acc = 0.;
    for (int n=0; n<pm.length(); n++){
      t = (int)(pm.t(n)*Fs) + t_shift;      
      if ((t < Nsig) && (t > 0))
	acc += sig.a_no_check(t);
    }
    if (fabs(acc) > max_acc){
      max_acc = fabs(acc);
      best_shift = t_shift;
    }
  }
  
  if (pm_debug_print) cout << "\tbest shift = " << best_shift << " samples \n";
  float fbs = (float) best_shift / Fs;
  for (int n=0; n<pm.length(); n++)
    pm.t(n) = pm.t(n) + fbs;
}


static void compute_E(EST_Wave &sig, OGI_Track &E, EST_Features &g){
  // actually power
  float guessF0 = g.F("guess_F0");
  float pppw = g.F("periods_per_Ewin");   // pitch periods per window
  int winsz = (int)(pppw * sig.sample_rate()/guessF0);
  int num_frames= (int)ceil(g.F("Eest_ppm")*sig.length()/(float)winsz); 
  E.resize(num_frames, 1); 
  E.fill_time(winsz/(float)(g.F("Eest_ppm")*sig.sample_rate()));

  EST_FVector frame;
  int window_start,pos, k;  
  EST_WindowFunc *wf =  EST_Window::creator("hamming");
  
  for (k=0; k < E.num_frames(); k++) {
    pos = (int)(E.t(k) * sig.sample_rate() + 0.5);
    window_start = pos - winsz/2;
    EST_Window::window_signal(sig, wf, window_start, winsz, frame, 1);
    s2p(frame, E.a(k));
  }
}

static void s2p(EST_FVector &frame, float &power){
  int i;
  power = 0.0;
  for (i=0; i < frame.length(); i++)
    power += frame(i)*frame(i);
  power /= frame.length();
}

static void max_limit_curve(OGI_Track &E){
  double meanE = E.mean(0);
  int k;
  for (k=0; k<E.num_frames(); k++)
    E.a(k,0) = min(meanE, E.a(k,0));
}
  
   
static void design_ma_filter(int N, EST_DVector &b){
  b.resize(N);
  for (int k=0; k<N; k++)
    b[k] = 1.0/N;
}

   
#define EXIT_THRESH 0.0001
static void cluster_2class(OGI_Track &fv, float Vbias, EST_FList &B, EST_FList &E){
  
  int j,k,Nv,Nuv;
  double prevCv,Cv=0.0,prevCuv,Cuv=0.0,du,dv;
  fv.resize(EST_CURRENT, fv.num_channels()+1);
  fv.set_channel_name("smoothed_energy",0);
  fv.set_channel_name("vuv",1);

  for (k=0; k<fv.num_frames(); k++){
    Cv = max(Cv,fv(k,0));
  }
  
  for (j=0; j<MAX_ITER; j++){
    for (k=0; k<fv.num_frames(); k++){
      dv = fv(k,0) - Cv;
      du = fv(k,0) - Cuv;
      fv(k,1) = (float)(Vbias*(dv*dv) < (du*du));
    }

    prevCv = Cv;
    prevCuv = Cuv;
    Cv = 0; Cuv = 0; Nv=0; Nuv=0;
    for (k=0; k<fv.num_frames(); k++){
      if ((int) fv(k,1)){
	Cv  += fv(k,0);
	Nv++;
      }
      else{
	Cuv += fv(k,0);
	Nuv++;
      }
    }
    Cv/=(float)Nv;
    Cuv/=(float)Nuv;

    if (fabs(Cv - prevCv) < EXIT_THRESH)
      break;
  }

  int wasVOICED = (int) fv(0,1);
  if (wasVOICED)
    B.append(0);
  
  for (k=0; k<fv.num_frames(); k++){
    if (((int) fv(k,1)) && (!wasVOICED))
      B.append(fv.t(k));
    else if ((!((int) fv(k,1))) && wasVOICED)
      E.append(fv.t(k));
    
    wasVOICED = (int) fv(k,1);
  }

  if (wasVOICED){
    E.append(fv.t(fv.num_frames()));
  }
}


static void find_vpm( EST_Wave &lx, EST_Wave &lxdiff, EST_FList &Blist, EST_FList &Elist, EST_IList &pm_cand){
  //     minima of Lx derivative between positive-going
  //         zero crossings of Lx in voiced regions
  
  EST_IList PZClist;
  int B,E,min_indx;
  int n,m;
  pm_cand.clear();
  for (n = 0; n < Blist.length(); n++){
    B = (int) (Blist.nth(n) * lx.sample_rate());
    E = (int) (Elist.nth(n) * lx.sample_rate());

    // safety
    B = max(B, 0); 
    E = min(E, lx.length()); 

    find_pgzc(lx, B, E, PZClist);
    PZClist.prepend(B);  // add beginning of voiced section

    OGI_Track tmpt;
    IList2Track(PZClist, lx.sample_rate(), tmpt);    

    if (pm_debug_dump) 
      tmpt.save("pgzc.est");

    for (m=0; m<PZClist.length()-1; m++){
      min_indx = find_min(lxdiff, PZClist.nth(m),  PZClist.nth(m+1));
      pm_cand.append(min_indx);
    }
  }
}

static void find_pgzc(EST_Wave &x, int B, int E, EST_IList &pgzc){
  pgzc.clear();
  int k;
  for (k=B+1; k<E; k++){
    if ((x(k-1) <= 0) && (x(k) > 0)){
      pgzc.append(k);
    }
  }
}


static int find_min(EST_Wave &x, int B, int E){
  int k;
  int min_i = B;
  short minval=32767;
  for (k=B; k<E; k++){
    if (x(k) < minval){
      min_i = k;
      minval = x(k);
    }
  }
  return min_i;
}



static bool rm_suspicious_pm(EST_Wave &lx, EST_IList &pmList, EST_Features &g){

  EST_IList suspicious;
  int prevpm=0,nextpm=0;
  int T0_min = (int) (lx.sample_rate()/(g.F("guess_F0")*g.F("F0_range_factor")));
  int T0_max = (int) (lx.sample_rate()/(g.F("guess_F0")/g.F("F0_range_factor")));

  ////
  //// stray?
  ////
  EST_IList rmlist;
  for (int k=0; k<pmList.length(); k++){
    prevpm = ((k==0) ? -100000 : pmList.nth(k-1));
    nextpm = ((k==pmList.length()-1) ?  pmList.last() + 10000 :  pmList.nth(k+1));
    
    // neither neighb is < T0_max away    
    if (((pmList.nth(k) - prevpm) > T0_max) && ((nextpm - pmList.nth(k)) > T0_max))
      rmlist.append(k);  // remove it
  }
  purge_pm_list(pmList, rmlist);
  

  ////
  //// double?
  ////
  for (int k=0; k<pmList.length(); k++){
    prevpm = ((k==0) ? -100000 : pmList.nth(k-1));
    nextpm = ((k==pmList.length()-1) ?  pmList.last() + 10000 :  pmList.nth(k+1));
    
    // nearest neighb is < T0_min away
    if (((pmList.nth(k) - prevpm) < T0_min) || ((nextpm - pmList.nth(k)) < T0_min)){
      suspicious.append(k);
    }
  }
  bool retval = (bool)(suspicious.length());

  // consider pairs of pmarks that are too close together
  int k=0;
  rmlist.clear();
  while (k<suspicious.length()){
    EST_IList pair;
    pair.clear();
    int first_of_pair = pmList.nth(suspicious.nth(k));
    if(k<suspicious.length()){
      pair.append(suspicious.nth(k));
      k++;
    }
    if((k<suspicious.length()) 
       && (pmList.nth(suspicious.nth(k))-first_of_pair < T0_min)){
      pair.append(suspicious.nth(k));
      k++;
    }

    if (pair.length() > 1){
      // measure: closeness to 1/2 way between pmarks just before and just after pair   
      int best_sc = 10000;
      int best_ix = 0;
      prevpm =  ((pair.first()==0) ? pmList.nth(0)-T0_min : pmList.nth(pair.first()-1));
      nextpm =  ((pair.last()==pmList.length()-1) ? pmList.last()+T0_min : pmList.nth(pair.last()+1));

      for (int l=0; l<pair.length(); l++){
	int sc = abs((pmList.nth(pair.nth(l)) - prevpm) - (nextpm - pmList.nth(pair.nth(l))));
	if (sc < best_sc){
	  best_sc = sc;
	  best_ix = l;
	}
      }
      
      // reduce each pair to 1 pmark
      for (int l=0; l<pair.length(); l++){
	if (l != best_ix){
	  rmlist.append(pair.nth(l));
	}
      }
    }
  }  
  purge_pm_list(pmList, rmlist);

  return retval;
}

static void purge_pm_list(EST_IList &pm, EST_IList &rmlist){
  // in pm, remove INDICES listed in rmlist
  EST_Litem *p;
  for (p=rmlist.tail(); p != 0; p = prev(p)){
    pm.remove_nth(rmlist(p));
  }
}


static void IList2Track(EST_IList &pmList, float Fs, OGI_Track &pm){
  int k;
  pm.resize(pmList.length(), 0);
  for ( k=0; k<pmList.length(); k++){
    pm.t(k) = pmList.nth(k)/Fs;
  }
}


static void parse_init_params(LISP params, EST_Features &In){
  LISP plisp=NIL;

  // Get params from Scheme layer
  In.set("lx_path",  wstrdup(get_param_str("lx_path", params, ".")));
  In.set("sp_path",  wstrdup(get_param_str("sp_path", params, ".")));
  In.set("lx_ext",   wstrdup(get_param_str("lx_ext",  params, ".lar")));
  In.set("sp_ext",   wstrdup(get_param_str("sp_ext",  params, ".wav")));
  In.set("in_path",  wstrdup(get_param_str("in_path", params, ".")));
  In.set("out_path", wstrdup(get_param_str("out_path",params, ".")));
  In.set("out_ext",  wstrdup(get_param_str("out_ext", params, ".pmv")));
  In.set("lx_sign",  wstrdup(get_param_str("lx_sign", params, "up_is_closed_glottis")));
  In.set("sex",      wstrdup(get_param_str("sex",     params, "male")));

  // NEED SPEECH_TOOLS PATCH....
  plisp = get_param_lisp("pre_filt1", params, NIL);
  if (plisp != NIL){
    In.set("pre_filt1", siod_sprint(plisp));
  }

  plisp = get_param_lisp("vuv_filt", params, NIL);
  if (plisp != NIL){
    In.set("vuv_filt", siod_sprint(plisp));
  }

  plisp = get_param_lisp("diff_filt", params, NIL);
  if (plisp != NIL){
    In.set("diff_filt", siod_sprint(plisp));
  }


  // other float parameters - no defaults to set
  LISP l;
  EST_String thing;
  for (l=params; l !=0; l=cdr(l)){
    thing = get_c_string(car(car(l)));
    if (!((In.present(thing)) || thing == "infile_list")){ 
      In.set(thing, get_c_float(car(cdr(car((l))))));
    }
  }

  // these are used in the rough LPC analysis for getting rid of lx-pmark delay
  In.set("ola_window_type","trapezoid");
  In.set("pre_emphasis",0.96);
  In.set("lpc_anl_frame_len",3.0);
  In.set("T0_UV_thresh", In.F("F0_range_factor") / In.F("guess_F0"));
  In.set("T0_UV_pm",0.01);

}
