/****************************<--OGI-->*************************************/
/*                                                                        */
/*             Center for Spoken Language Understanding                   */
/*        Oregon Graduate Institute of Science & Technology               */
/*                         Portland, OR USA                               */
/*                        Copyright (c) 1999                              */
/*                                                                        */
/*      This module is not part of the CSTR/University of Edinburgh       */
/*               release of the Festival TTS system.                      */
/*                                                                        */
/*  In addition to any conditions disclaimers below, please see the file  */
/*  "license_ogi_tts.txt" distributed with this software for information  */
/*  on usage and redistribution, and for a DISCLAIMER OF ALL WARRANTIES.  */
/*                                                                        */
/****************************<--OGI-->*************************************/
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include "festival.h"
#include "resLPC.h"
#include "resLPC_io.h"
#include "OGI_Track.h"
#include "OGI_WaveChunk.h"
#include "OGIdbase.h"
#include "OGIdbase_scm.h"
#include "OGIresLPC_db.h"



static void append_join_info(EST_Item *newUnit, OGIresLPC_SRC *srcdata, EST_Track *pm_buf);
static void find_join_offsets( OGIresLPC_SRC *srcdata, EST_Track *pm_buf, EST_Item *newUnit, int *Ladjust, int *Radjust);
static void update_srcseg(EST_Utterance *u, EST_Item *dbUnit, int *Ladjust, int *Radjust, float Fs);
static bool join_is_voiced(OGI_Track &Lpm, EST_Track &Rpm, float LsegE);


//
//  Residual-excited LPC top level synthesis routine
//
void resLPC_concat(
	EST_Utterance *u,      
	OGIresLPC_SRC *srcdata // container for all retrieved db units
	){

  OGIdbase *dbase = get_db();  // ptr to current unit inventory
  OGIresLPC_db *udb = (OGIresLPC_db *) dbase->udata;
  
  // temp buffers for db access
  EST_Wave  *exc_buf = NULL;
  EST_Track *pm_buf = NULL;
  EST_Track *lpc_buf = NULL;
  int excB,excE,indx;
  float pm_offset;
  int Ladjust, Radjust;

  EST_Item *newUnit;

  srcdata->Fs = dbase->Fs;
  srcdata->exc.set_sample_rate((int) srcdata->Fs);

  // Check for existence of dbUnit relation in utt
  if (!u->relation_present("dbUnit")){
    cerr << "OGIresLPC:  no dbUnit stream present in utterance" << endl;
    festival_error();
  }

  // create SrcSeg relation
  //    - carries characteristics (durations, etc)  of concatenated segments
  u->create_relation("SrcSeg");

  // Loop through the units
  for (newUnit=u->relation("dbUnit")->head(); newUnit != 0; newUnit=next(newUnit)){
    //  Items in the dbUnit relation have the following features:
    //    - "db_indx" into dbase
    //    - "isNatNbL" - is the prev unit this unit's natural neighbor?
    //    - "isSegBndL" - is the prev unit in the same segment as this unit?
    //    -  a set of daughters
    //            each with at least the feature "dur"

    // retreive resLPC unit from database -- excitation, pitchmarks, and lpc  
    indx = newUnit->f("db_indx");
    udb->get_dbUnit(*dbase, indx, &exc_buf, &excB, &excE, &pm_buf, &lpc_buf);

    // Update join info 
    append_join_info(newUnit, srcdata, pm_buf);

    // Adjust join based on join type
    find_join_offsets(srcdata, pm_buf, newUnit, &Ladjust, &Radjust);

    // Update segment info in SrcSeg
    update_srcseg(u, newUnit, &Ladjust, &Radjust, srcdata->Fs);

    // record join point (in samples)
    newUnit->set("exc", srcdata->exc.length() + Ladjust);

    // time offset for pitchmarks
    pm_offset = (srcdata->exc.length()+Ladjust - Radjust)/srcdata->Fs;
    
    // copy excitation WAV 
    srcdata->exc.adjust_last_segE(Ladjust);
    srcdata->exc.append_chunk(*exc_buf, excB+Radjust, excE, newUnit->I("isNatNbL"));

    // copy PITCHMARKS
    srcdata->pm.offset_append(*pm_buf, pm_offset);

    // copy LPC COEFFS
    srcdata->lpc.offset_append(*lpc_buf, pm_offset);
  }
}

static void append_join_info(
			     EST_Item *newUnit,
			     OGIresLPC_SRC *srcdata,
			     EST_Track *pm_buf
			     ){
  //  adds features to newUnit
  float LsegE = srcdata->exc.length()/srcdata->Fs; // sec

  // If this is the 1st unit of utt
  if (same_item(newUnit->first(), newUnit)){
    newUnit->set("isFirstUnit",1);
    newUnit->set("m/a","abut");
    newUnit->set("pm", 0);
    newUnit->set("lpcpm", 0);
  } 
    
  // 
  // decide whether to abut or merge
  //
  if (newUnit->I("isNatNbL")){  // if prev unit is this one's natural neighbor
    newUnit->set("m/a","abut");
  }
  else {  
    if (join_is_voiced(srcdata->pm, *pm_buf, LsegE))  // voiced at join point
      newUnit->set("m/a","merge");
    else                               // unvoiced at join point
      newUnit->set("m/a","abut");
  }
  
  // first of new unit
  newUnit->set("pm", srcdata->pm.num_frames());
  newUnit->set("lpcpm", srcdata->lpc.num_frames());
}



static void find_join_offsets(
			      OGIresLPC_SRC *srcdata,
			      EST_Track *pm_buf,
			      EST_Item *newUnit,
			      int *Ladjust,  // samples
			      int *Radjust   // samples
			      ){
  *Ladjust = 0;
  *Radjust = 0;
  
  if (newUnit->I("isFirstUnit",0)){
    newUnit->set("exc", 0);
    return;
  }

  int LsegE  = srcdata->exc.length();  // end of left segment (samp)

  if (newUnit->S("m/a") == "merge"){
    // If MERGE, put 1st pmark of new unit at one pitch period past end
    //  of last one in the utterance so far.      
    float LpmLast = srcdata->pm.last_t(); // last pm of unit to left of join
    float RpmFirst = pm_buf->t(0);        // first pm of unit to right  
    float T0_join = srcdata->pm.last_T0(); // fund period just before the join (sec)
    *Ladjust = (int)((LpmLast  + T0_join/2.)*srcdata->Fs - LsegE);
    *Radjust = (int)((RpmFirst - T0_join/2.)*srcdata->Fs);
  }
  else {
    // else just abut as-is with no shift
    *Ladjust = 0;
    *Radjust = 0;
  }
}

static void update_srcseg(EST_Utterance *u, EST_Item *this_unit, int *Ladj_int, int *Radj_int, float Fs){

  EST_Item *segitem,*newseg;
  EST_Item scopy;
  EST_Relation *SrcSegRel = u->relation("SrcSeg");

  float Ladjust = (*Ladj_int)/Fs;
  float Radjust = (*Radj_int)/Fs;

  // adjust duration of segment left of the join
  if (Ladjust != 0.0){
    segitem = SrcSegRel->last();
    Ladjust = max(Ladjust, -(segitem->F("dur")-0.001)); // don't let it take more than is there
    segitem->set("dur", segitem->F("dur") + Ladjust); 
  }

  // get ptr to 1st daughter seg of the new dbUnit
  newseg = daughter1(this_unit);

  if (newseg!=0){
    // adjust 1st newseg (right) duration 
    Radjust = min(Radjust, newseg->F("dur")-0.001); // don't let it take more than is there
    newseg->set("dur", newseg->F("dur") - Radjust); 
  }

  // if join is not a segment boundary, just add to dur of prev one
  //   (note: assumes other features remain unchanged)
  if (!(this_unit->I("isSegBndL") || this_unit->I("isFirstUnit",0))){
    if (newseg!=0){
      segitem = SrcSegRel->last();
      segitem->set("dur", segitem->F("dur") + newseg->F("dur")); 
      newseg = next_sibling(newseg);
    }
    else
      cerr << "   **newseg=NULL! \n";
  }

  // append new item(s) for remaining segments in unit
  for ( ; newseg!=0; newseg=next_sibling(newseg)){
    // append to SrcSeg relation
    SrcSegRel->append(newseg);    
  }

  // send these back,  possibly changed
  *Ladj_int = (int)(Ladjust*Fs + 0.5);
  *Radj_int = (int)(Radjust*Fs + 0.5);
}

			  
  
			  

static bool join_is_voiced(OGI_Track &Lpm, EST_Track &Rpm, float LsegE){
  // NOTE: this method will give poor results if we have a unit with < 2 pitchmarks!
  //   This will be fixed if we decide to store pmarks marked as V or UV -
  //   then we can just look at a V/UV field in the Track.
  EST_Features *gl=get_resLPC_params();
  float T0_UV_thresh = 0.0;
  T0_UV_thresh = gl->F("T0_UV_thresh");

  if ((Rpm.num_frames() > 1) && (Lpm.num_frames() > 1)){ 
    float L_T0 = Lpm.last_T0();
    float R_T0 = Rpm.t(1) - Rpm.t(0);

    if ((Rpm.t(0) < T0_UV_thresh) 
	&& ((LsegE - Lpm.last_t()) < T0_UV_thresh)){
      if (L_T0 != OGI_UNDEFINED){
	if ((L_T0 < T0_UV_thresh) && (R_T0 < T0_UV_thresh))
	  return TRUE;
      }
    }
  }
  return FALSE;
}
