/****************************<--OGI-->*************************************/
/*                                                                        */
/*             Center for Spoken Language Understanding                   */
/*        Oregon Graduate Institute of Science & Technology               */
/*                         Portland, OR USA                               */
/*                        Copyright (c) 1999                              */
/*                                                                        */
/*      This module is not part of the CSTR/University of Edinburgh       */
/*               release of the Festival TTS system.                      */
/*                                                                        */
/*  In addition to any conditions disclaimers below, please see the file  */
/*  "license_ogi_tts.txt" distributed with this software for information  */
/*  on usage and redistribution, and for a DISCLAIMER OF ALL WARRANTIES.  */
/*                                                                        */
/****************************<--OGI-->*************************************/
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include "EST.h"
#include "resLPC.h"
#include "resLPC_io.h"
#include "resLPC_mod.h"
#include "OGI_Track.h"
#include "OGI_TimeWarp.h"
#include "OGI_WaveChunk.h"
#include "OGI_Buffer.h"
#include "OGI_Macro.h"
#include "OGIduration.h"
#include "OGIpitch.h"
#include "Phone.h"
#include "resLPC_pmark.h"

static void add_mod_channels(OGIresLPC_SRC *src, OGIresLPC_MOD *mod);
static void mod_silences( EST_Utterance *u, OGIresLPC_SRC *src, OGIresLPC_MOD *mod);
static void make_T0_src_contour(OGIresLPC_SRC *src);
static void make_output_pmarks_direct( OGIresLPC_SRC *src, OGIresLPC_MOD *mod);
static void ramp_sil_energy(OGI_Track &lpc, float t_in, int direction);
static void make_beta_contour( EST_Utterance *u, OGIresLPC_SRC *src, OGIresLPC_MOD *mod);
static void make_default_beta_contour(OGIresLPC_SRC *src);
static void make_output_pmarks_beta( OGIresLPC_SRC *src, OGIresLPC_MOD *mod);
static void update_srcseg_pmark_info( EST_Utterance *u, OGIresLPC_SRC *src);
static int get_smooth_region(LISP smoothTable, EST_Item *Lph, EST_Item *Rph);
static void match_track_joins( EST_Utterance *u, OGI_Track &srclpc, int chB, int Nch, LISP smoothTable, bool doCrossPh, bool doMatch);
static void match_joins_power( EST_Utterance *u, OGIresLPC_SRC *src );
static void match_joins_spectra( EST_Utterance *u, OGIresLPC_SRC *src);
static EST_Item *first_srcseg_in_unit(EST_Utterance *u, EST_Item *dbUnit);
static EST_Item *last_srcseg_in_unit(EST_Utterance *u, EST_Item *dbUnit);
static void apply_ext_gain( EST_Utterance *u, OGIresLPC_SRC *src, OGI_TimeWarp &warp );

//
// "direct" method:  interpolate within target F0 contour to
//    get target pitchmarks, use T0 of original pmarks in UV regions
//
// "softF0" method:  interpolate within target F0 contour to
//    get pitch modification factor beta for each src pmark,
//    then (optionally) smooth this and find output pmarks by
//    beta*T0_src.  This will not give desirable behavior when
//    the time duration of the target is much longer than the
//    source and there are many pitch fluctuations (e.g., vibrato).
//
void resLPC_mod( 
		EST_Utterance *u, 
		OGIresLPC_SRC *src, 
		OGIresLPC_MOD *mod){

  EST_Features *gl = get_resLPC_params();

  // Add needed channels into Tracks
  add_mod_channels(src,mod);

  // Generate time warping function from durations
  if (gl->S("dur_method") == "natural")
    make_default_time_warp(u, mod->warp);
  else
    make_time_warp(u, mod->warp);

  // time stretch silences if needed
  mod_silences(u, src, mod);

  // Make voicing decision 
  pmark_V_UV(src->pm, src->Bvsect, src->Evsect, src->exc.length(), src->Fs,  *gl);

  // Add UV pmarks
  make_UVpmarks(src->pm, src->Bvsect, src->Evsect, src->exc.length(), src->Fs,  *gl, 1);

  // Bookkeeping for pmark locations relative to SrcSeg
  update_srcseg_pmark_info(u, src);

  // Get F0 targets
  get_target_F0(u, mod->F0targ, mod->warp, gl->F("F0_default"));

  // smooth spectral features
  match_joins_spectra(u, src);
  
  // smooth power
  match_joins_power(u, src);

  // external gain adjustment
  apply_ext_gain(u, src, mod->warp);

  // pitch modification
  if (gl->S("pitch_method") == "soft")
    make_beta_contour(u, src, mod); //  pitch mod factor
  else if (gl->S("pitch_method") == "natural")
    make_default_beta_contour(src); //  pitch mod factors == 1.0
  else // "direct"
    make_T0_src_contour(src);    // calc T0 for each pitchmark
    
  // create synthesis pitchmarks
  if ((gl->S("pitch_method") == "soft") || (gl->S("pitch_method") == "natural"))
    make_output_pmarks_beta(src, mod);
  else
    make_output_pmarks_direct(src, mod);
}


//
// "none" method:  don't modify duration or F0 at all
//
void resLPC_NOmod( 
		   EST_Utterance *u, 
		   OGIresLPC_SRC *src, 
		   OGIresLPC_MOD *mod){
  
  EST_Features *gl = get_resLPC_params();

  // Add needed channels into Tracks
  add_mod_channels(src,mod);

  // Generate time warping function from durations
  make_default_time_warp(u, mod->warp);

  // Make voicing decision 
  pmark_V_UV(src->pm, src->Bvsect, src->Evsect, src->exc.length(), src->Fs,  *gl);

  // Add UV pmarks
  make_UVpmarks(src->pm, src->Bvsect, src->Evsect, src->exc.length(), src->Fs,  *gl, 1);

  // Bookkeeping for pmark locations relative to SrcSeg
  update_srcseg_pmark_info(u, src);

   // Make "beta" pitch mod factors == 1.0
  make_default_beta_contour(src);

  // create synthesis pitchmarks
  make_output_pmarks_beta(src, mod);

}


static void add_mod_channels(OGIresLPC_SRC *src, 
			     OGIresLPC_MOD *mod){
  
  int ch;
  src->pm.resize(EST_ALL, 3);

  ch = 0;
  src->pm.set_channel_name("v/uv",ch++);
  src->pm.set_channel_name("T0_src",ch++);
  src->pm.set_channel_name("beta",ch++);
  
  mod->pm.resize(EST_ALL, 3);
  ch = 0;
  mod->pm.set_channel_name("lpc_map",ch++);
  mod->pm.set_channel_name("exc_map",ch++);
  mod->pm.set_channel_name("v/uv",ch++);
  
}



//
// Add zero samples at joins to duration-stretch silences
//   (assumes phone-based duration targets)
//
static void mod_silences(
			 EST_Utterance *u,
			 OGIresLPC_SRC *src,
			 OGIresLPC_MOD *mod
			 ){
  int i;
  float expand,t_insert;
  EST_Item *useg;
  EST_Item *sseg;

  // Loop thru segments
  for (i=0, useg=u->relation("Segment")->head(), sseg=u->relation("SrcSeg")->head(); 
       sseg != 0; 
       useg=next(useg), sseg=next(sseg), i++) {

    if (useg == 0)
      break;  // if different # of segs (already warned above)

    if (ph_is_silence(useg->name())){
      if (mod->warp.modfactor(i) > 1.00) {

	// how much time to insert
	expand = mod->warp.targ_dur(i) - mod->warp.src_dur(i); 

	// set insert point and ramp energy to 0 in initial/final silences
	if (i==0){                  // initial silence 
	  t_insert = 0.0;
	  ramp_sil_energy(src->lpc, mod->warp.src_end(i), -1);
	}
	if (i==mod->warp.Nseg()-1){ // final silence
	  t_insert =  mod->warp.src_end(i);
	  ramp_sil_energy(src->lpc, mod->warp.src_begin(i),  1);
	}	  
	else
	  t_insert =  mod->warp.src_begin(i) + mod->warp.src_dur(i)/2.;
	
	// modify the srcseg duration
	mod->warp.modify_src_dur(i, expand);
	sseg->set("dur", sseg->F("dur")+expand);
	     
	// offset speech pmarks
	src->pm.offset_t_after(t_insert, expand);
	
	// offset LPC pmarks
	src->lpc.offset_t_after(t_insert, expand);
	
	// insert zeros in wave
	t_insert = max(0.0, min((src->exc.length()-1)/src->Fs, t_insert)); 
	src->exc.insert_zeros((int)(t_insert*src->Fs), (int)(expand*src->Fs));
      }
    }
  }
}
static void ramp_sil_energy(OGI_Track &lpc, float t_in, int direction){
  
  int chE = lpc.channel_position("E");
  int B,E,k;
  float g,incr;

  if (direction == 1){ // down to end
    g = 0; // == gain of 1.0
    B = lpc.index_above(t_in);
    E = lpc.num_frames()-1;
  }
  else {               // down to beginning
    g = -10; // == near 0
    B = 0;
    E = lpc.index_below(t_in);
  }

  if (E > B){
    incr = -direction * 10/(float)(E-B);
    for (k=B; k<=E; k++){
      lpc.a(k,chE) = lpc.a(k,chE) + g;
      g += incr;
    }
  }
} 

//
// Generate T0_src for each pmark
//
static void make_T0_src_contour(OGIresLPC_SRC *src){
  int k;
  EST_Track T0_src;
  src->pm.sub_track(T0_src, 0, EST_ALL, src->pm.channel_position("T0_src"), 1);
  int Npm = src->pm.num_frames();

  for (k=0; k<Npm-1; k++){
    // source To contour
    T0_src(k) = src->pm.T0(k);
  }

  // last one
  if (Npm > 1){
    T0_src(Npm-1) = T0_src(Npm-2);
  }
}

//
// Generate T0_src and pitch mod factor (beta) for each pmark
//
static void make_beta_contour(
				 EST_Utterance *u,
				 OGIresLPC_SRC *src,
				 OGIresLPC_MOD *mod
				 ){

  int k;
  EST_Track T0_src;
  src->pm.sub_track(T0_src, 0, EST_ALL, src->pm.channel_position("T0_src"), 1);
  EST_Track beta;
  src->pm.sub_track(beta, 0, EST_ALL, src->pm.channel_position("beta"), 1);
  EST_Track v_uv;
  src->pm.sub_track(v_uv, 0, EST_ALL, src->pm.channel_position("v/uv"), 1);

  int Npm = src->pm.num_frames();


  for (k=0; k<Npm; k++){

    // source To contour
    T0_src(k) = src->pm.T0(k);

    // pitch mod factor
    if (v_uv(k) == _V_)
      beta(k) =  mod->F0targ.linterp(mod->warp.fwdmap(src->pm.t(k))) * T0_src(k);
    else
      beta(k) =  1.0;   // no pitch mod for unvoiced
  }

  // smooth voiced sections of beta() between join points
  EST_Features *gl=get_resLPC_params();
  int Bsmoothlen=gl->I("beta_smooth");
  int chBETA =  src->pm.channel_position("beta");
  
  int B=0,E=0;
  EST_Item *dbUnit=u->relation("dbUnit")->head();    
  for  (dbUnit=next(dbUnit); dbUnit != 0; dbUnit=next(dbUnit)){
    E = dbUnit->I("pm");
    
    if (dbUnit->I("isNatNbL")){
      // don't smooth yet if nat neighbor to prev
      continue;
    }
    else{      
      // do smoothing
      src->pm.smooth_ma(B, E-1, chBETA, 1, Bsmoothlen);
      
      // start new one
      B = E;
    }
  }
  src->pm.smooth_ma(B, src->pm.num_frames()-1, chBETA, 1, Bsmoothlen);

  // reset UV ones to 1.0 
  //  (need better way - still can get weird effects at voicing boundaries)
  for (k=0; k<Npm; k++){
    if (v_uv(k) != _V_)
      beta(k) =  1.0;
  }
}

//
// Generate an T0_src and pitch mod factor (beta) for each pmark
//
static void make_default_beta_contour(OGIresLPC_SRC *src){
  int k;
  EST_Track T0_src;
  src->pm.sub_track(T0_src, 0, EST_ALL, src->pm.channel_position("T0_src"), 1);
  EST_Track beta;
  src->pm.sub_track(beta, 0, EST_ALL, src->pm.channel_position("beta"), 1);
  
  for (k=0; k<src->pm.num_frames(); k++){
    T0_src(k) = src->pm.T0(k);
    beta(k) =  1.0;   // no pitch mod 
  }
}


static void update_srcseg_pmark_info(
				     EST_Utterance *u, 
				     OGIresLPC_SRC *src){

  int k_pm=0; 
  int k_lpcpm=0;
  int Npm=src->pm.num_frames();
  int Nlpcpm=src->lpc.num_frames();
  float src_end=0.0;
  EST_Item *sseg;
  
  // Loop thru SrcSeg, record 1st pmark of each segment
  for (sseg=u->relation("SrcSeg")->head(); sseg!=0; sseg=next(sseg)){
    
    src_end  += sseg->F("dur");

    // pmarks
    sseg->set("pm", k_pm);
    while ((k_pm < Npm) && (src->pm.t(k_pm) < src_end )){
      k_pm++;
    }
  
    // lpc pmarks
    sseg->set("lpcpm", k_lpcpm);
    while ( (k_lpcpm < Nlpcpm) && (src->lpc.t(k_lpcpm) < src_end )){
      k_lpcpm++;
    }
  }
}

//
// apply correction to match LSF trajectories at joins
//
static void match_joins_spectra(
				 EST_Utterance *u,
				 OGIresLPC_SRC *src
				 ){    
  EST_Features *gl = get_resLPC_params();
  if (gl->present("spectra_smooth")){
    LISP smoothTable = read_from_lstring(strintern(gl->S("spectra_smooth")));
    bool doCrossPh = (bool) gl->I("smooth_cross_ph_join");
    bool doMatch = TRUE;
    
    if (gl->S("spectra_match_or_replace") == "replace")
      doMatch = FALSE;
    
    int chLSF0 = src->lpc.channel_position("LSF_0");
    int Nlsf   = src->lpc.channel_position("LSF_N")-chLSF0+1;
    
    match_track_joins(u, src->lpc, chLSF0, Nlsf, smoothTable, doCrossPh, doMatch);
  }
}

//
// apply correction to match power at joins
//
static void match_joins_power(
			      EST_Utterance *u,
			      OGIresLPC_SRC *src
			      ){    
  EST_Features *gl = get_resLPC_params();

  if (gl->present("power_smooth")){
    LISP smoothTable = read_from_lstring(strintern(gl->S("power_smooth")));
    bool doCrossPh = (bool) gl->I("smooth_cross_ph_join");
    bool doMatch = TRUE;
    
    if (gl->S("power_match_or_replace") == "replace")
      doMatch = FALSE;
    
    match_track_joins(u, src->lpc, src->lpc.channel_position("E"), 1, 
		      smoothTable, doCrossPh, doMatch);
  }
}

static void match_track_joins(
			      EST_Utterance *u,
			      OGI_Track &srclpc,
			      int chB,  // begin channel in track
			      int Nch,  // number of channels to smooth
			      LISP smoothTable, // smoothing info
			      bool doCrossPh,
			      bool doMatch){    
  EST_Item  *Lph,*Rph;
  int smoothlen=0,B=0,M=0,E=0,nextpm=0;
  
  EST_Item *dbUnit=u->relation("dbUnit")->head();    

  // last one in the first unit
  Lph = last_srcseg_in_unit(u, dbUnit);

  for  (dbUnit=next(dbUnit); dbUnit != 0; dbUnit=next(dbUnit)){
    // first phone in this unit
    Rph = first_srcseg_in_unit(u, dbUnit);

    smoothlen = 0;
    if (!(dbUnit->I("isNatNbL"))){  // don't smooth if nat neighbor to prev

      // smooth if this is intra-phone join or if cross-phone smoothing enabled
      if (!(dbUnit->I("isSegBndL")) || doCrossPh){
	
	// iterate through lisp list of context features to match against phones
	smoothlen = get_smooth_region(smoothTable, Lph, Rph);
      }
    }
    
    // join point
    M = dbUnit->I("lpcpm");
    
    // begin point: limit to begin of unit and begin of segment the join is in
    B = max(prev(dbUnit)->I("lpcpm"), M - smoothlen);
    B = max(B, Lph->I("lpcpm"));
        
    // end point: limit to end of unit and end of segment the join is in
    if (next(dbUnit))
      nextpm = next(dbUnit)->I("lpcpm");
    else
      nextpm =  srclpc.num_frames();
    E = min(nextpm, M + smoothlen);
    if (next(Rph))
      E = min(E, next(Rph)->I("lpcpm")-1);
    
    // do smoothing
    if (doMatch)
      srclpc.linear_match(B, M, E, chB, Nch);   // adjustment of track
    else
      srclpc.linear_replace(B, M, E, chB, Nch); // replacement of track
        
    // last phone in this unit
    Lph = last_srcseg_in_unit(u, dbUnit);
  }
}


static int get_smooth_region(LISP smoothTable, EST_Item *Lph, EST_Item *Rph){
  EST_String phfeat,targ_featval;
  LISP l;
  int smoothlen = 0;

  for (l=smoothTable; l != NIL; l=cdr(l)){
    phfeat       = EST_String(wstrdup(get_c_string(car(car(l)))));
    targ_featval = EST_String(wstrdup(get_c_string(car(cdr(car(l))))));

    if ((targ_featval == ph_feat(Lph->name(), phfeat)) 
	&& (targ_featval == ph_feat(Rph->name(), phfeat))){
      smoothlen = get_c_int(car(cdr(cdr(car(l)))));
      break;
    }
  }
  return smoothlen;
}


//
// apply external gain modification
//
static void apply_ext_gain(
			   EST_Utterance *u,
			   OGIresLPC_SRC *src,
			   OGI_TimeWarp &warp
			   ){    

  if (!(u->relation_present("Gain")))
    return;

  OGI_SingleTrack Gtarg;
  EST_Item *targ;
  const EST_Relation *utt_targ = u->relation("Gain");

  // Exit if there are no gain targets given
  if (utt_targ->length() == 0)
    return;
  

  // prepend a target if the time of the 1st target is not at time=0
  if (utt_targ->first()->F("pos") != 0.0)
    Gtarg.prepend(0.0, 1.0);
  // loop through targets
  for (targ=utt_targ->first(); targ != 0; targ=next(targ)) {
    Gtarg.append(targ->F("pos"), targ->F("gain"));
  }  
  // Ensure there is a target at the end 
  if (utt_targ->last()->F("pos") < u->relation("Segment")->last()->F("end")){
    Gtarg.append(warp.last_targ_end(), utt_targ->last_leaf()->F("gain"));
  }

  // Loop through src lpc pitchmarks, map fwd to apply gain value
  int ch = src->lpc.channel_position("E");
  int N = src->lpc.num_frames();
  for (int k=0; k<N; k++){
    src->lpc.a(k, ch) += log(10e-10 + Gtarg.linterp(warp.fwdmap(src->lpc.t(k))));
  }
}



static void make_output_pmarks_direct(
				      OGIresLPC_SRC *src,
				      OGIresLPC_MOD *mod
				      ){  
  int n=0;
  int chT0_SRC = src->pm.channel_position("T0_src");
  int chEXCMAP = mod->pm.channel_position("exc_map");
  int chLPCMAP = mod->pm.channel_position("lpc_map");
  int chVUV    = mod->pm.channel_position("v/uv");
  int exc_map;

  EST_Track v_uv;
  src->pm.sub_track(v_uv, 0, EST_ALL, src->pm.channel_position("v/uv"), 1);
  
  OGI_Track *srcpm = &(src->pm);
  OGI_Track *modpm = &(mod->pm);
  float modT0, ta, ts=0.0;

  // generate pitchmarks through end of last target Segment
  n = 0;
  modpm->append_t_buf_start();
  while (ts < mod->warp.last_targ_end()){

    // (buffered) append to modified pmarks
    modpm->append_t_buf(ts);

    // mapping from ts to src time ta
    ta = mod->warp.invmap(ts);

    // map to nearest exc pmark indx
    modpm->a(n, chEXCMAP) = srcpm->nearest_indx_f(ta);

    // map to nearest LPC pmark indx
    modpm->a(n, chLPCMAP) = src->lpc.nearest_indx_f(ta);

    // modified T0: use original T0 in UV regions
    exc_map = (int)(modpm->a(n, chEXCMAP) + 0.5);
    if (v_uv(exc_map) == _V_){
      modpm->a(n, chVUV) = _V_;
      modT0 = 1/mod->F0targ.linterp(ts);
    }
    else {
      modpm->a(n, chVUV) = _UV_;
      modT0 = srcpm->nearest(ta, chT0_SRC);
    }

    ts += modT0;
    n++;
  }
  modpm->append_t_buf_finish();
}



static void make_output_pmarks_beta(
				    OGIresLPC_SRC *src,
				    OGIresLPC_MOD *mod
				    ){  
  int n=0;
  int chT0_SRC = src->pm.channel_position("T0_src");
  int chBETA   = src->pm.channel_position("beta");
  int chEXCMAP = mod->pm.channel_position("exc_map");
  int chLPCMAP = mod->pm.channel_position("lpc_map");
  int chVUV    = mod->pm.channel_position("v/uv");
  
  EST_Track v_uv;
  src->pm.sub_track(v_uv, 0, EST_ALL, src->pm.channel_position("v/uv"), 1);

  OGI_Track *srcpm = &(src->pm);
  OGI_Track *modpm = &(mod->pm);
  float modT0, ta, ts=0.0;
  int exc_map;

  // generate pitchmarks through end of last target Segment
  while (ts < mod->warp.last_targ_end()){

    // append to targ_pm Track, return indx # of pmark
    n = modpm->append_t(ts);

    // mapping from ts to src time ta
    ta = mod->warp.invmap(ts);

    // map to nearest wav pmark indx
    modpm->a(n, chEXCMAP) = srcpm->nearest_indx_f(ta);

    // map to nearest LPC pmark indx
    modpm->a(n, chLPCMAP) = src->lpc.nearest_indx_f(ta);

    // keep track of voicing
    exc_map = (int)(modpm->a(n, chEXCMAP) + 0.5);
    if (v_uv(exc_map) == _V_)
      modpm->a(n, chVUV) = _V_;
    else 
      modpm->a(n, chVUV) = _UV_;

    // modified T0: (T0_src nearest to ta)*(beta interpolated at ta)
    modT0 = srcpm->nearest(ta, chT0_SRC) / srcpm->interp(ta, chBETA);
    ts += modT0;

  }
}


static EST_Item *first_srcseg_in_unit(EST_Utterance *u, EST_Item *dbUnit){
  int join = dbUnit->I("lpcpm");
  EST_Item *sseg=u->relation("SrcSeg")->head();

  for (; sseg!=0; sseg=next(sseg)){
    if (sseg->I("lpcpm") > join)
      break;
  }

  if (sseg == 0)
    return u->relation("SrcSeg")->last();
  if (prev(sseg))
    return prev(sseg);  
  return sseg;
}

static EST_Item *last_srcseg_in_unit(EST_Utterance *u, EST_Item *dbUnit){
  int join;
  if (next(dbUnit))
    join = next(dbUnit)->I("lpcpm");
  else
    return u->relation("SrcSeg")->last();

  EST_Item *sseg=u->relation("SrcSeg")->head();

  for (; sseg!=0; sseg=next(sseg)){
    if (sseg->I("lpcpm") >= join)
      break;
  }
  if (prev(sseg))
    return prev(sseg);  
  return sseg;
}



// tied to these specific definitions of "v/uv" and _UV_
//  this is a very slow way to do it
void rm_unvoiced_pmarks(OGI_Track *pm){
  OGI_Track *newpm = new OGI_Track;
  int chV = pm->channel_position("v/uv");
  if (chV < 0)
    return;

  int n,k;

  newpm->resize(pm->num_frames(), pm->num_channels());
  for (n=0,k=0; n < pm->num_frames(); n++){
    if (pm->a(n, chV) != _UV_){
      newpm->copy_frame_in(k, *pm, n);
      k++;
    }
  }

  delete pm;
  pm = newpm;
}
